#include <gtk/gtk.h>
#include <m17n.h>
#include <stdio.h>
#include <string.h>

static void
entry_cb (GtkWidget *entry, gpointer data)
{
  const gchar *text = gtk_entry_get_text (GTK_ENTRY (entry));
  MSymbol lang, name, variable, key;
  MPlist *plist = mplist ();

  lang = g_object_get_data (G_OBJECT (entry), "mim_lang");
  name = g_object_get_data (G_OBJECT (entry), "mim_name");
  variable = g_object_get_data (G_OBJECT (entry), "mim_variable");
  key = g_object_get_data (G_OBJECT (entry), "mim_key");

  if (key == Msymbol)
    {
      mplist_add (plist, Msymbol, msymbol (text));
      minput_config_variable (lang, name, variable, plist);
    }
  else if (key == Mtext)
    {
      MText *mt;

      mt = mtext_from_data (text, strlen (text), MTEXT_FORMAT_UTF_8);
      mplist_add (plist, Mtext, mt);
      minput_config_variable (lang, name, variable, plist);
      m17n_object_unref (mt);
    }
  else if (key == Minteger)
    {
      int i;
      gchar buf[32];

      if (sscanf (text, "%d", &i) == 1)
	{
	  mplist_add (plist, Minteger, (void *) i);
	    minput_config_variable (lang, name, variable, plist);
	}
      else
	{
	  printf ("Invalid value\n");
	}
      sprintf (buf, "%d", (int) mplist_value (plist));
      gtk_entry_set_text (GTK_ENTRY (entry), buf);
      gtk_editable_set_position (GTK_EDITABLE (entry), -1);
    }
  m17n_object_unref (plist);
}

static void
combo_cb (GtkWidget *combo, gpointer data)
{
  gchar *text = gtk_combo_box_get_active_text (GTK_COMBO_BOX (combo));
  MSymbol lang, name, variable, key;
  MPlist *plist = mplist ();

  lang = g_object_get_data (G_OBJECT (combo), "mim_lang");
  name = g_object_get_data (G_OBJECT (combo), "mim_name");
  variable = g_object_get_data (G_OBJECT (combo), "mim_variable");
  key = g_object_get_data (G_OBJECT (combo), "mim_key");

  if (key == Msymbol)
    {
      mplist_add (plist, Msymbol, msymbol (text));
      minput_config_variable (lang, name, variable, plist);
    }
  else if (key == Mtext)
    {
      MText *mt;

      mt = mtext_from_data (text, strlen (text), MTEXT_FORMAT_UTF_8);
      mplist_add (plist, Mtext, mt);
      minput_config_variable (lang, name, variable, plist);
      m17n_object_unref (mt);
    }
  else if (key == Minteger)
    {
      int i;
      gchar buf[32];

      if (sscanf (text, "%d", &i) == 1)
	{
	  mplist_add (plist, Minteger, (void *) i);
	  minput_config_variable (lang, name, variable, plist);
	}
      else
	{
	  printf ("Invalid value\n");
	}
    }
  m17n_object_unref (plist);
}

static void
spin_cb (GtkWidget *spin, gpointer data)
{
  MSymbol lang, name, variable;
  MPlist *plist = mplist ();

  lang = g_object_get_data (G_OBJECT (spin), "mim_lang");
  name = g_object_get_data (G_OBJECT (spin), "mim_name");
  variable = g_object_get_data (G_OBJECT (spin), "mim_variable");

  mplist_add
    (plist, Minteger,
     (void *) gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (spin)));
  minput_config_variable (lang, name, variable, plist);
  m17n_object_unref (plist);
}

void
create_variable_entries (GtkWidget *frame, GtkTooltips *tip,
			 MSymbol lang, MSymbol name)
{
  GtkWidget *table;
  MPlist *plist;
  gint row;

  plist = minput_get_variable (lang, name, Mnil);
  if (! plist)
    {
      GtkWidget *message;

      message = gtk_label_new ("No variables for this method");
      gtk_container_add (GTK_CONTAINER (frame), message);
      return;
    }

  /*
   * plist == ((variable description status value [valid-value ...])
   *           (variable description status value [valid-value ...])
   *           ...)
   */
  table = gtk_table_new (2, mplist_length (plist), FALSE);
  gtk_container_add (GTK_CONTAINER (frame), table);

  for (row = 0; plist && mplist_key (plist) == Mplist;
       row++, plist = mplist_next (plist))
    {
      GtkWidget *label, *widget;
      MPlist *pl, *data;
      MSymbol variable, key;
      void *value;
      gchar *desc;

      pl = mplist_value (plist);
      variable = mplist_value (pl);
      label = gtk_label_new (msymbol_name (variable));
      gtk_table_attach_defaults (GTK_TABLE (table), label, 0, 1, row, row + 1);

      pl = mplist_next (pl);
      if (mplist_key (pl) == Mtext)
	/* Fixme : Assuming the return value is in UTF-8 */
	desc = mtext_data (mplist_value (pl), NULL, NULL, NULL, NULL);
      else
	desc = NULL;

      pl = mplist_next (mplist_next (pl)); /* ignore status */
      key = mplist_key (pl);
      value = mplist_value (pl);
      pl = mplist_next (pl);

      if (key == Msymbol)
	{
	  if (mplist_key (pl) == Msymbol)
	    {
	      int i, nth = -1;

	      widget = gtk_combo_box_new_text ();
	      for (i = 0; pl && mplist_key (pl) == Msymbol;
		   pl = mplist_next (pl), i++)
		{
		  if (mplist_value (pl) == value)
		    nth = i;
		  gtk_combo_box_append_text (GTK_COMBO_BOX (widget),
					     msymbol_name (mplist_value (pl)));
		}
	      if (nth != -1)
		gtk_combo_box_set_active (GTK_COMBO_BOX (widget), nth);
	      g_signal_connect (G_OBJECT (widget), "changed",
				G_CALLBACK (combo_cb), NULL);
	    }
	  else
	    {
	      widget = gtk_entry_new ();
	      gtk_entry_set_text (GTK_ENTRY (widget), msymbol_name (value));
	      gtk_editable_set_editable (GTK_EDITABLE (widget), TRUE);
	      g_signal_connect (G_OBJECT (widget), "activate",
				G_CALLBACK (entry_cb), NULL);
	    }
	}
      
      else if (key == Mtext)
	{
	  if (mplist_key (pl) == Mtext)
	    {
	      int i, nth = -1;

	      widget = gtk_combo_box_new_text ();
	      for (i = 0; pl && mplist_key (pl) == Mtext;
		   pl = mplist_next (pl), i++)
		{
		  if (! mtext_cmp (mplist_value (pl), value))
		    nth = i;
		  /* Fixme : Assuming the return value is in UTF-8 */
		  gtk_combo_box_append_text
		    (GTK_COMBO_BOX (widget),
		     mtext_data (mplist_value (pl), NULL, NULL, NULL, NULL));
		}
	      if (nth != -1)
		gtk_combo_box_set_active (GTK_COMBO_BOX (widget), nth);
	      g_signal_connect (G_OBJECT (widget), "changed",
				G_CALLBACK (combo_cb), NULL);
	    }
	  else
	    {
	      widget = gtk_entry_new ();
	      /* Fixme : Assuming the return value is in UTF-8 */
	      gtk_entry_set_text (GTK_ENTRY (widget),
				  mtext_data (value, NULL, NULL, NULL, NULL));
	      gtk_editable_set_editable (GTK_EDITABLE (widget), TRUE);
	      g_signal_connect (G_OBJECT (widget), "activate",
				G_CALLBACK (entry_cb), NULL);
	    }
	}

      else if (key == Minteger)
	{
	  if (mplist_key (pl) == Minteger)
	    {
	      int i, nth = -1;

	      widget = gtk_combo_box_new_text ();
	      for (i = 0; pl && mplist_key (pl) == Minteger;
		   pl = mplist_next (pl), i++)
		{
		  gchar buf[32];

		  if (mplist_value (pl) == value)
		    nth = i;
		  snprintf (buf, 31, "%d", (int) mplist_value (pl));
		  gtk_combo_box_append_text (GTK_COMBO_BOX (widget), buf);
		}
	      if (nth != -1)
		gtk_combo_box_set_active (GTK_COMBO_BOX (widget), nth);
	      g_signal_connect (G_OBJECT (widget), "changed",
				G_CALLBACK (combo_cb), NULL);
	    }
	  else if (mplist_key (pl) == Mplist)
	    {
	      GtkAdjustment *adj;
	      gdouble lower, upper;

	      pl = mplist_value (pl);
	      lower = (gdouble) (int) mplist_value (pl);
	      pl = mplist_next (pl);
	      upper = (gdouble) (int) mplist_value (pl);
	      adj = (GtkAdjustment *)
		gtk_adjustment_new ((gdouble) (int) value, (gdouble) lower,
				    (gdouble) upper, 1.0, 10.0, 0);
	      widget = gtk_spin_button_new (adj, 0, 0);
	      gtk_spin_button_set_numeric (GTK_SPIN_BUTTON (widget), TRUE);
	      gtk_spin_button_set_update_policy (GTK_SPIN_BUTTON (widget),
						 GTK_UPDATE_IF_VALID);
	      g_signal_connect (G_OBJECT (widget), "value_changed",
				G_CALLBACK (spin_cb), NULL);
	    }
	  else
	    {
	      gchar buf[32];

	      widget = gtk_entry_new ();
	      snprintf (buf, 31, "%d", (int) value);
	      gtk_entry_set_text (GTK_ENTRY (widget), buf);
	      gtk_editable_set_editable (GTK_EDITABLE (widget), TRUE);
	      g_signal_connect (G_OBJECT (widget), "activate",
				G_CALLBACK (entry_cb), NULL);
	    }
	}

      else
	{
	  widget = gtk_entry_new ();
	  gtk_entry_set_text (GTK_ENTRY (widget), "???");
	  gtk_editable_set_editable (GTK_EDITABLE (widget), TRUE);
	  g_signal_connect (G_OBJECT (widget), "activate",
			    G_CALLBACK (entry_cb), NULL);
	}

      gtk_tooltips_set_tip (tip, widget, desc, NULL);
      g_object_set_data (G_OBJECT (widget), "mim_lang", lang);
      g_object_set_data (G_OBJECT (widget), "mim_name", name);
      g_object_set_data (G_OBJECT (widget), "mim_variable", variable);
      g_object_set_data (G_OBJECT (widget), "mim_key", key);
      gtk_table_attach_defaults	(GTK_TABLE (table), widget,
				 1, 2, row, row + 1);
    }
}
